local UIF_MAJOR, UIF_MINOR = "AceGUI-Factory-1.0", 1
local UIF, oldminor = LibStub:NewLibrary(UIF_MAJOR, UIF_MINOR)

if not UIF then return end
local AceGUI = LibStub("AceGUI-3.0")

UIF.factories = UIF.factories or {}

local Embed

local defaultFont = GameFontNormal:GetFont()

UIF.defaults = {
	font = defaultFont,
	fontSize = 12,
	fontColor = {177/255, 219/255, 255/255},
	fontFlags = "none",
	borderColor = {0.2,0.3,0.37,1},
	bgColor = {0.11,0.16,0.19,1},
	
	MainTitle = {
		fontSize = 35,
		fontColor = {0.4, 0.6, 0.8},
		fontFlags = "OUTLINE"
	},
	Level1Title = {
		fontSize = 25,
		fontColor = {18/255, 157/255, 143/255},
		fontFlags = "OUTLINE"
	},
	NormalText = {
		fontSize = 16,
		fontColor = {177/255, 219/255, 255/255},
	},
	ListText = {
		fontSize = 16,
		fontColor = {0.96, 0.93, 0.49},
	},
	Frame = {
		fontSize = 30,
		bgColor = {0.11,0.16,0.19,0.8},
	},
	DropdownGroup = {
	},
	Dropdown = {
		itemFontSize = 18,
	},
	TreeGroup = {
		fontColor = {0.4, 0.6, 0.8},		
	},
	TabGroup = {	
		fontSize = 25,
		bgColor = {0.11,0.16,0.19,1},
		borderColor = {0.2,0.3,0.37,1},
	},
	TabGroup2 = {	
		fontSize = 25,
		fontFlags = "OUTLINE",
		bgColor = {40/255, 45/255, 70/255},
		borderColor = {81/255, 96/255, 106/255},
		fontColor = {207/255, 218/255, 85/255},
	},
	InlineGroup = {
		fontFlags = "OUTLINE",
		fontSize = 22,
	},
	InlineGroup2 = {
		bgColor = {40/255, 45/255, 70/255},
		borderColor = {81/255, 96/255, 106/255},
		fontColor = {207/255, 218/255, 85/255},
		fontSize = 22,
	},
	Button = {
	},
	Slider = {		
	},
	CheckBox = {
	},
}

function UIF:GetFactory(name, obj)
	assert(type(obj) == "table" or not obj, "Invalid factory object, it must be either nil or a table")
	
	if self.factories[name] then return self.factories[name] end
	
	if obj then
		assert(obj.s, "Factory object must contain the \"s\" table containing the default settings to be used")
		local s = obj.s
		for opt, value in pairs(self.defaults) do
			if type(value) == "table" then
				if (s[opt]) then
					assert(type(s[opt]) == "table", ("s[%q] must be a %s"):format(opt, type(value)))
					for subOpt, subValue in pairs(value) do
						if s[opt][subOpt] then
							assert(type(s[opt][subOpt]) == type(subValue), ("s[%q][%q] must be a %s"):format(opt, subOpt, type(subValue)))
						end
						s[opt][subOpt] = s[opt][subOpt] or self.defaults[opt][subOpt]
					end
				end
				s[opt] = s[opt] or self.defaults[opt]
			else
				if s[opt] then
					assert(type(s[opt]) == type(value), ("s[%q] must be a %s"):format(opt, type(value)))
				end
				s[opt] = s[opt] or value
			end
		end
		
		for opt, _ in pairs(s) do
			assert(self.defaults[opt], ("The setting %q is not a vailid setting"):format(opt))
		end
	end
	
	obj = obj or {}
	
	obj.name = name
	
	Embed(obj)
	
	self.factories[name] = obj
	
	return obj
end

local function OnMouseEnter(widget, event)
	local title, text = widget:GetUserData("enterTitle"), widget:GetUserData("enterText")
	GameTooltip:SetOwner(widget.frame)
	if title then
		GameTooltip:AddLine(title)
	end
	GameTooltip:AddLine(text)
	GameTooltip:Show()
end

local function OnMouseLeave(wdiget, event)
	GameTooltip:Hide()
end

local function SetMouseOver(self, widget, title, text)
	widget:SetUserData("enterTitle", title)
	widget:SetUserData("enterText", text)
	widget:SetCallback("OnEnter", OnMouseEnter)
	widget:SetCallback("OnLeave", OnMouseLeave)
end

local function GetLabel(lType, self, text, color, size, font, flags, interactive)
	if text then	assert(type(text) == "string" or type(text) == "number", ("%s-GuiFactory. %q must be a string or number"):format(self.name, "text")) end
	if color then 	assert(type(color) 	== "table" and color[1] and #color > 2 and 
		type(color[1]) == "number" and
		type(color[2]) == "number" and
		type(color[3]) == "number", ("%s-GuiFactory. %q must be a numeric indexed table with at least 3 values, each of numeric type"):format(self.name, "color")) end
	if size then 	assert(type(size) 	== "number", ("%s-GuiFactory. %q must be a number"):format(self.name, "size")) end
	if font then	assert(type(font) 	== "string", ("%s-GuiFactory. %q must be a string"):format(self.name, "font")) end
	if flags then 	assert(type(flags) 	== "string", ("%s-GuiFactory. %q must be a string"):format(self.name, "flags")) end
	
	local s = self.s[lType]
	
	color = color or s.fontColor or self.s.fontColor
	local label
	if interactive then
		label = AceGUI:Create("InteractiveLabel")
	else
		label = AceGUI:Create("Label")
	end
	
	label:SetImage([[Interface\Icons\Achievement_Dungeon_Ulduar77_25man]], 0, 1, 0, 1)
	label:SetImageSize(0, 0)
	label:SetImage(nil)
	label:SetFont(font or s.font or self.s.font, size or s.fontSize or self.s.fontSize, flags or s.fontFlags or self.s.fontFlags)
	label:SetColor(unpack(color))
	label:SetText(text)
	label:SetRelativeWidth(1)
	return label
end

--------------------
-- Labels

local function MainTitleLabel(self, text, color, size, font, flags, interactive)
	return GetLabel("MainTitle", self, text, color, size, font, flags, interactive)
end

local function Level1Title(self, text, color, size, font, flags, interactive)
	return GetLabel("Level1Title", self, text, color, size, font, flags, interactive)
end

local function ListText(self, text, color, size, font, flags, interactive)
	return GetLabel("ListText", self, text, color, size, font, flags, interactive)
end

local function NormalText(self, text, color, size, font, flags, interactive)
	return GetLabel("NormalText", self, text, color, size, font, flags, interactive)
end

---------------------
-- Frame

local function Frame(self, text)
	local s = self.s["Frame"]
	
	local group = AceGUI:Create("UIF-Frame")
	group:SetTitleFont(s.font or self.s.font, s.fontSize or self.s.fontSize, s.fontFlags or self.s.fontFlags)
	group:SetTitleFontColor(s.fontColor or self.s.fontColor)
	group:SetBorderColor(s.borderColor or self.s.borderColor)
	group:SetBackColor(s.bgColor or self.s.bgColor)
	group:SetTitle(text)
	
	local button = self:Button("")
	
	group:SetButton(button)
	
	return group
end

---------------------
-- Dropdown group

local function DropdownGroup(self, text)
	local s = self.s["DropdownGroup"]
	
	local group = AceGUI:Create("UIF-DropdownGroup")
	group:SetFullWidth(true)
	group:SetFont(s.font or self.s.font, s.fontSize or self.s.fontSize, s.fontFlags or self.s.fontFlags)
	group:SetFontColor(s.fontColor or self.s.fontColor)
	group:SetBorderColor(s.borderColor or self.s.borderColor)
	
	local drop = self:Dropdown()
	group:SetDropdownWidget(drop)
	return group
end

----------------------
-- Tree group

local function TreeGroup(self)
	local s = self.s["TreeGroup"]
	
	local group = AceGUI:Create("UIF-TreeGroup")
	group:SetFullWidth(true)
	group:SetFont(s.font or self.s.font, s.fontSize or self.s.fontSize, s.fontFlags or self.s.fontFlags)
	group:SetFontColor(s.fontColor or self.s.fontColor)
	group:SetSubFont(s.subFont or self.s.font, s.subFontSize or self.s.fontSize, s.subFontFlags or self.s.fontFlags)
	group:SetSubColor(s.subColor or self.s.fontColor)
	group:SetBorderColor(s.borderColor or self.s.borderColor)
	return group
end

----------------------
-- Tab group

local function TGroup(tType, self, text)
	if text then	assert(type(text)	== "string", ("%s-GuiFactory. %q must be a string"):format(self.name, "text")) end
	
	local s = self.s[tType]
	
	local group = AceGUI:Create("UIF-TabGroup")
	group:SetTitle(text)
	group:SetFont(s.font or self.s.font, s.fontSize or self.s.fontSize, s.fontFlags or self.s.fontFlags)
	group:SetTabFont(s.tabFont or self.s.font, s.tabFontSize or self.s.fontSize, s.tabFontFlags or self.s.fontFlags)
	group:SetFontColor(s.fontColor or self.s.fontColor)
	
	group:SetBorderColor(s.borderColor or self.s.borderColor)
	group:SetBackColor(s.bgColor)
	
	group:SetFullWidth(true)
	group:SetLayout("Flow")
	RWUIFTAB = group
	return group
end

local function TabGroup(self, text)
	return TGroup("TabGroup", self, text)
end

local function TabGroup2(self, text)
	return TGroup("TabGroup2", self, text)
end

----------------------
-- Inline group

local function IGroup(gType, self, text, bgColor, borderColor)
	if text then	assert(type(text)	== "string", ("%s-GuiFactory. %q must be a string"):format(self.name, "text")) end
	if bgColor then 	assert(type(bgColor) 	== "table" and #bgColor > 2 and 
		type(bgColor[1]) == "number" and
		type(bgColor[2]) == "number" and
		type(bgColor[3]) == "number", ("%s-GuiFactory. %q must be a numeric indexed table with at least 3 values, each of numeric type"):format(self.name, "bgColor")) end
	if borderColor then 	assert(type(borderColor) 	== "table" and #borderColor > 2 and 
		type(borderColor[1]) == "number" and
		type(borderColor[2]) == "number" and
		type(borderColor[3]) == "number", ("%s-GuiFactory. %q must be a numeric indexed table with at least 3 values, each of numeric type"):format(self.name, "borderColor")) end
	
	local s = self.s[gType]
	
	local group = AceGUI:Create("UIF-InlineGroup")
	group:SetLayout("Flow")
	group:SetFullWidth(true)
	group:SetTitle(text)
	group:SetFont(s.font or self.s.font, s.fontSize or self.s.fontSize, s.fontFlags or self.s.fontFlags)
	group:SetFontColor(s.fontColor or self.s.fontColor)
	group:SetBGColor(bgColor or s.bgColor or self.s.bgColor)
	group:SetBorderColor(borderColor or s.borderColor or self.s.borderColor)
	
	return group
end

local function InlineGroup(self, text, bgColor, borderColor)
	return IGroup("InlineGroup", self, text, bgColor, borderColor)
end

local function InlineGroup2(self, text, bgColor, borderColor)
	return IGroup("InlineGroup2", self, text, bgColor, borderColor)
end

----------------------
-- Changing widgets

local function UpdateTrigger(widget, event, ...)
	local self = widget:GetUserData("self")	
	local db = widget:GetUserData("db")
	local key = widget:GetUserData("key")
	local func = widget:GetUserData("extraCallback")
	local TriggerGlobalUpdate = widget:GetUserData("TriggerGlobalUpdate")
	
	if db then
		if select("#", ...) > 1 then
			for i = 1, select("#", ...) do
				db[key] = db[key] or {}
				if type(db[key]) ~= "table" then
					db[key] = {}
				end
				db[key][i] = select(i, ...)
			end
		else
			db[key] = ...
		end
	end
	
	if func then func(widget, event, ...) end
	
	if widget.SetValue then
		widget:SetValue(...)
	end
	if widget.SetColor then
		widget:SetColor(...)
	end
	if TriggerGlobalUpdate then
		if self.valuesChangedCallback then self.valuesChangedCallback(self) end
	end
end

local function ConfirmPupup(widget, event, ...)
	if not StaticPopupDialogs["AEGUIFACTORY_CONFIRM_DIALOG"] then
		StaticPopupDialogs["AEGUIFACTORY_CONFIRM_DIALOG"] = {}
	end
	local t = StaticPopupDialogs["AEGUIFACTORY_CONFIRM_DIALOG"]
	for k in pairs(t) do
		t[k] = nil
	end
	t.text = "Are you sure?"
	t.button1 = ACCEPT
	t.button2 = CANCEL
	local dialog, oldstrata
	t.OnAccept = function()
		UpdateTrigger(widget, event)
		if dialog and oldstrata then
			dialog:SetFrameStrata(oldstrata)
		end
	end
	t.OnCancel = function()
		if dialog and oldstrata then
			dialog:SetFrameStrata(oldstrata)
		end
	end
	for i = 1, select('#', ...) do
		t[i] = select(i, ...) or false
	end
	t.timeout = 0
	t.whileDead = 1
	t.hideOnEscape = 1

	dialog = StaticPopup_Show("AEGUIFACTORY_CONFIRM_DIALOG")
	if dialog then
		oldstrata = dialog:GetFrameStrata()
		dialog:SetFrameStrata("TOOLTIP")
	end
end

local function SetGUIOption(widget, event, ...)
	local self = widget:GetUserData("self")
	local db = widget:GetUserData("db")
	local key = widget:GetUserData("key")
	local UpdateOnMouseUp = widget:GetUserData("UpdateOnMouseUP")
	local TriggerGlobalUpdate = widget:GetUserData("TriggerGlobalUpdate")
	local func = widget:GetUserData("extraCallback")
	local confirm = widget:GetUserData("confirm")
	
	if confirm then ConfirmPupup(widget, event) return end
	
	if db then	
		if select("#", ...) > 1 then
			for i = 1, select("#", ...) do
				db[key] = db[key] or {}
				if type(db[key]) ~= "table" then
					db[key] = {}
				end
				db[key][i] = select(i, ...)
			end
		else
			db[key] = ...
		end
	end
	
	if widget.SetValue then
		widget:SetValue(...)
	end
	if widget.SetColor then
		widget:SetColor(...)
	end
	
	if func then func(widget, event, ...) end
	if TriggerGlobalUpdate then
		if not UpdateOnMouseUp and self.valuesChangedCallback then self.valuesChangedCallback(self) end
	end
end

local function Slider(self, text, db, key, min, max, step, callback)
	assert(text and type(text) == "string", ("%s-GuiFactory. %q must be a string"):format(self.name, "text"))
	assert(db and type(db) == "table", ("%s-GuiFactory. %q must be a table"):format(self.name, "db"))
	assert(key, ("%s-GuiFactory. %q must be supplied"):format(self.name, "key"))
	assert(db[key], ("%s-GuiFactory. %q does not contain %q"):format(self.name, "db", key))
	if callback then assert(type(callback) == "function", ("%s-GuiFactory. %q must be a function"):format(self.name, "callback")) end
	if min then assert(type(min) == "number", ("%s-GuiFactory. %q must be a number"):format(self.name, "min")) end
	if max then assert(type(max) == "number", ("%s-GuiFactory. %q must be a number"):format(self.name, "max")) end
	if step then assert(type(step) == "number", ("%s-GuiFactory. %q must be a number"):format(self.name, "step")) end
	
	local s = self.s["Slider"]	
	
	local slider = AceGUI:Create("UIF-Slider")
	slider:SetLabel(text)
	slider:SetFont(s.font or self.s.font, s.fontSize or self.s.fontSize, s.fontFlags or self.s.fontFlags)
	slider:SetFontColor(s.fontColor or self.s.fontColor)
	slider:SetSliderValues(min or 0, max or 100, step or 1)
	slider:SetValue(db[key])
	
	slider:SetUserData("db", db)
	slider:SetUserData("key", key)
	slider:SetUserData("extraCallback", callback)
	slider:SetUserData("self", self)
	slider:SetUserData("UpdateOnMouseUP", true)
	slider:SetCallback("OnValueChanged", SetGUIOption)
	slider:SetCallback("OnMouseUp", UpdateTrigger)
	
	return slider
end

local function CheckBox(self, text, db, key, callback)
	assert(text and type(text) == "string", ("%s-GuiFactory. %q must be a string"):format(self.name, "text"))
	assert(db and type(db) == "table", ("%s-GuiFactory. %q must be a table"):format(self.name, "db"))
	assert(key, ("%s-GuiFactory. %q must be supplied"):format(self.name, "key"))
	if callback then assert(type(callback) == "function", ("%s-GuiFactory. %q must be a function"):format(self.name, "callback")) end
	
	local s = self.s["CheckBox"]
	
	local check = AceGUI:Create("UIF-CheckBox")
	check:SetFullWidth(true)
	check:SetLabel(text)
	check:SetFont(s.font or  self.s.font, s.fontSize or self.s.fontSize, s.fontFlags or self.s.fontFlags)
	check:SetFontColor(s.fontColor or self.s.fontColor)
	check:SetValue(db[key])
	
	check:SetUserData("db", db)
	check:SetUserData("key", key)
	check:SetUserData("extraCallback", callback)
	check:SetUserData("self", self)
	check:SetCallback("OnValueChanged", SetGUIOption)
	
	return check
end

local function ColorSelect(self, text, db, key, callback)
	assert(text and type(text) == "string", ("%s-GuiFactory. %q must be a string"):format(self.name, "text"))
	assert(db and type(db) == "table", ("%s-GuiFactory. %q must be a table"):format(self.name, "db"))
	assert(key, ("%s-GuiFactory. %q must be supplied"):format(self.name, "key"))
	if callback then assert(type(callback) == "function", ("%s-GuiFactory. %q must be a function"):format(self.name, "callback")) end
	
	local s = self.s["CheckBox"]
	
	local color = AceGUI:Create("UIF-ColorPicker")
	color:SetLabel(text)
	color:SetFont(s.font or self.s.font, s.fontSize or self.s.fontSize, s.fontFlags or self.s.fontFlags)
	color:SetFontColor(s.fontColor or self.s.fontColor)
	if db[key] and type(db[key]) == "table" then
		color:SetColor(unpack(db[key]))
	end
	
	color:SetUserData("db", db)
	color:SetUserData("key", key)
	color:SetUserData("extraCallback", callback)
	color:SetUserData("self", self)
	color:SetCallback("OnValueChanged", SetGUIOption)
	color:SetUserData("UpdateOnMouseUP", true)
	color:SetCallback("OnValueConfirmed", UpdateTrigger)	
	
	return color
end

-----------------------
-- Other

local function Button(self, text, callback, font, size, flags, color)
	assert(text and type(text) == "string", ("%s-GuiFactory. %q must be a string"):format(self.name, "text"))
	if color then 	assert(type(color) 	== "table" and color[1] and #color > 2 and 
		type(color[1]) == "number" and
		type(color[2]) == "number" and
		type(color[3]) == "number", ("%s-GuiFactory. %q must be a numeric indexed table with at least 3 values, each of numeric type"):format(self.name, "color")) end
	if size then 	assert(type(size) 	== "number", ("%s-GuiFactory. %q must be a number"):format(self.name, "size")) end
	if font then	assert(type(font) 	== "string", ("%s-GuiFactory. %q must be a string"):format(self.name, "font")) end
	if flags then 	assert(type(flags) 	== "string", ("%s-GuiFactory. %q must be a string"):format(self.name, "flags")) end
	
	local s = self.s["Button"]
	
	color = color or s.fontColor or self.s.fontColor
	
	local button = AceGUI:Create("UIF-Button")
	button:SetText(text)
	button:SetFont(font or s.font or self.s.font, size or s.fontSize or self.s.fontSize, flags or s.fontFlags or self.s.fontFlags)
	button:SetFontColor(color)
	
	button:SetUserData("self", self)
	button:SetUserData("extraCallback", callback)
	button:SetCallback("OnClick", SetGUIOption)	
	
	button:SetWidth(button.text:GetStringWidth() + 30)
	
	return button
end

----------------------
-- Dropdown

local function Dropdown(self, text)
	local s = self.s["Dropdown"]

	local drop = AceGUI:Create("UIF-Dropdown")
	drop:SetLabel(text)
	drop:SetItemFont(s.itemFont or self.s.font, s.itemFontSize or self.s.fontSize, s.itemFontFlags or self.s.fontFlags)
	drop:SetItemFontColor(s.itemFontColor or self.s.fontColor)
	drop:SetFont(s.font or self.s.font, s.fontSize or self.s.fontSize, s.fontFlags or self.s.fontFlags)
	drop:SetFontColor(s.fontColor or self.s.fontColor)
	
	return drop
end

local function SetValuesChangedCallback(self, func)
	assert(type(func) == "function", "The Values changed callback must be a function")
	self.valuesChangedCallback = func
end

local embeds = {
	MainTitleLabel = MainTitleLabel,
	Level1Title = Level1Title,
	ListText = ListText,
	NormalText = NormalText,
	Frame = Frame,
	DropdownGroup = DropdownGroup,
	TreeGroup = TreeGroup,
	TabGroup = TabGroup,
	TabGroup2 = TabGroup2,
	InlineGroup = InlineGroup,
	InlineGroup2 = InlineGroup2,
	Slider = Slider,
	Button = Button,
	CheckBox = CheckBox,
	ColorSelect = ColorSelect,
	Dropdown = Dropdown,
	SetValuesChangedCallback = SetValuesChangedCallback,
	SetMouseOver = SetMouseOver,
}

function Embed(target)
	for i, emb in pairs(embeds) do
		target[i] = emb
	end
end

for fName, factory in pairs(UIF.factories) do
	Embed(factory)
end